package edu.wpi.banmqp.shimmer;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;

import android.annotation.SuppressLint;
import android.bluetooth.BluetoothSocket;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import edu.wpi.banmqp.basestation.MESSAGE_TYPE;
import edu.wpi.banmqp.shimmer.commands.CommandFactory;
import edu.wpi.banmqp.shimmer.commands.learned.ILearnedCommand;
import edu.wpi.banmqp.shimmer.commands.learned.LearnedCommand;
import edu.wpi.banmqp.shimmer.commands.learned.ParamMapping;
import edu.wpi.banmqp.shimmer.commands.learned.ReturnMapping;
import edu.wpi.banmqp.shimmer.data.ValueMapping;
import edu.wpi.banmqp.shimmer.sensor.ISensor;
import edu.wpi.banmqp.shimmer.sensor.SensorFactory;
import edu.wpi.banmqp.shimmer.sensor.SensorMap;
import edu.wpi.banmqp.util.BANLog;

/**
 * The Class BasicShimmerMote.
 * 
 */
public class BasicShimmerMote implements IShimmer {

	/** The sock. */
	BluetoothSocket sock;

	/** The in. */
	private BufferedInputStream in;

	private OutputStream out;

	private SensorMap sensorMap;

	private SequenceNumber sequenceNumber;

	private boolean cascadeLearn;

	private boolean cts;

	private Queue<byte[]> messageQueue;

	private boolean isStreaming, wasStreaming;

	/**
	 * Instantiates a new basic shimmer mote.
	 * 
	 * @param sock the sock
	 * @param sensorType the sensor type
	 */
	public BasicShimmerMote(BluetoothSocket sock) {
		this.sock = sock;
		sequenceNumber = new SequenceNumber();
		sensorMap = new SensorMap();
		cascadeLearn = false;
		cts = true;
		messageQueue = new LinkedList<byte[]>();
		isStreaming = wasStreaming = false;
		try {
			// Connect if we aren't already
			while (!sock.isConnected()) {
				BANLog.d("MOTE_CONNECT", "Trying to connect to "
						+ sock.getRemoteDevice().getAddress());
				sock.connect();
			}
			BANLog.d("MOTE_CONNECT", "Connected to "
					+ sock.getRemoteDevice().getAddress());
			// Construct buffered streams
			in = new BufferedInputStream(sock.getInputStream());
			out = sock.getOutputStream();


		} catch (IOException e) {
			e.printStackTrace();
		}
	}



	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.banmqp.shimmer.IShimmer#getData(int)
	 */
	@Override
	public byte[] getData(int numBytesToGet) {
		byte[] buff = new byte[numBytesToGet];
		try {
			while (in.available() < numBytesToGet) {
			}
			in.read(buff, 0, numBytesToGet);
		} catch (IOException e) {
			e.printStackTrace();
			BANLog.e("READ_ERROR", "Unable to read " + numBytesToGet + " from "
					+ sock.getRemoteDevice().getAddress());
		}
		return buff;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.banmqp.shimmer.IShimmer#bytesAvailible()
	 */
	@Override
	public int bytesAvailible() {
		try {
			return in.available();
		} catch (IOException e) {
			e.printStackTrace();
			BANLog.e("READ_ERROR",
					"Socket does not appear to be connected, or the inputstream is null");
			return -1;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.banmqp.shimmer.IShimmer#startStreaming()
	 */
	@Override
	public void startStreaming() {


	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.banmqp.shimmer.IShimmer#stopStreaming()
	 */
	@Override
	public void stopStreaming() {

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.banmqp.shimmer.IShimmer#getBluetoothAddr()
	 */
	@Override
	public String getBluetoothAddr() {
		return sock.getRemoteDevice().getAddress();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.banmqp.shimmer.IShimmer#isConnected()
	 */
	@Override
	public boolean isConnected() {
		return sock.isConnected();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.banmqp.shimmer.IShimmer#getSensorIDs()
	 */
	@Override
	public Integer[] getSensorIDs() {
		Set<Integer> idset = sensorMap.keySet();
		return idset.toArray(new Integer[0]);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.banmqp.shimmer.IShimmer#getSensors()
	 */
	@Override
	public ISensor[] getSensors() {
		Integer[] ids = getSensorIDs();
		ISensor[] sensors = new ISensor[ids.length];
		for (int i = 0; i < ids.length; i++) {
			sensors[i] = sensorMap.get(ids[i]);
		}

		return sensors;
	}

	@Override
	public ISensor getSensorById(int id) {
		return sensorMap.get(id);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.banmqp.shimmer.IShimmer#getSensorNames()
	 */
	@Override
	public String[] getSensorNames() {
		ISensor[] sensors = getSensors();
		String[] sensorNames = new String[sensors.length];
		for (int i = 0; i < sensors.length; i++) {
			sensorNames[i] = sensors[i].toString();
		}
		return sensorNames;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.banmqp.shimmer.IShimmer#processPacket()
	 */
	@Override
	public synchronized void processPacket(Handler h) {
		try {
			BANLog.d("PROCESS_PACKET", "Waiting for packet");
			// Get packet size
			ByteBuffer sizeBuff = ByteBuffer.allocate(2);
			sizeBuff.order(ByteOrder.BIG_ENDIAN);
			sizeBuff.put((byte) in.read());
			sizeBuff.put((byte) in.read());
			int numBytes = sizeBuff.getShort(0);
			BANLog.d("PROCESS_PACKET", "Packet size " + numBytes);
			// Throw out the sequence number
			in.read();
			in.read();
			int messageID = in.read();
			BANLog.d("PROCESS_PACKET", "Message ID: " + messageID);
			int payloadSize = numBytes - 5;
			byte[] packetData = new byte[payloadSize];
			int payloadRead = 0;
			// Since in.read(byte[] buff) doesn't seem to block on a bufferedinputstream, read this in bytewise to make
			// sure we get everything
			for(int i = 0; i < payloadSize; i++){
				packetData[i] = (byte) in.read();
				payloadRead++;
			}
			BANLog.d("PROCESS_PACKET", "Expected payload " + payloadSize + " actually got " + payloadRead);
			BANLog.d("PROCESS_PACKET", "RAW DATA WITHOUT HEADER: " + Arrays.toString(packetData));
			cts = true;
			switch (messageID) {
			case MESSAGE_ID.SENSOR_INQUIRY:
				processSensorInquiry(h, packetData);
				break;
			case MESSAGE_ID.COMMAND_INQUIRY:
				processCommandInquiry(h, packetData);
				break;
			case MESSAGE_ID.COMMAND_PARAMS_INQUIRY:
				processCommandParamsInquiry(h, packetData);
				break;
			case MESSAGE_ID.COMMAND_RETURNS_INQUIRY:
				processCommandReturnsInquiry(h, packetData);
				break;
			case MESSAGE_ID.DATA_INQUIRY:
				processDataInquiry(h, packetData);
				break;
			case MESSAGE_ID.MOTE_DATA:
				processMoteData(h, packetData);
				break;
				// TODO: add handling of learned command responses
			case MESSAGE_ID.LEARNED_COMMAND:
				processLearnedCommand(h, packetData);
			default: 
				break;

			}
		} catch (IOException e) {
			BANLog.e("PROCESS_PACKET", "Error reading packet");
		}
		return;
	}

	@Override
	public void sendMoteData(){
		queueSendBytes(CommandFactory.MoteData(sequenceNumber));
		isStreaming = !isStreaming;
		// we need to waste a bit of time so the mote can actually start/stop
		for(int i = 0; i < 5; i++){
			queueSendBytes(new byte[0]);
		}
	}

	@Override
	public void sendSensorInquiry() {
		queueSendBytes(CommandFactory.SensorInquiry(sequenceNumber));

	}

	@Override
	public void sendSensorInquiry(boolean cascade) {
		if(cascade){
			BANLog.d("CASCADE_LEARN", "Cascade learn started, sending sensor inquiry");
			cascadeLearn = true;
		}
		queueSendBytes(CommandFactory.SensorInquiry(sequenceNumber));

	}

	@Override
	public void sendCommandInquiry(int sensorID) {
		queueSendBytes(CommandFactory.CommandInquiry(sensorID, sequenceNumber));

	}

	@Override
	public void sendCommandParamsInquiry(int SensorID, int commandID) {
		queueSendBytes(CommandFactory.CommandParamsInquiry(SensorID, commandID,
				sequenceNumber));

	}

	@Override
	public void sendCommandReturnsInquiry(int sensorID, int commandID) {
		queueSendBytes(CommandFactory.CommandReturnsInquiry(sensorID, commandID,
				sequenceNumber));

	}

	@Override
	public void sendDataInquiry(int sensorID) {
		queueSendBytes(CommandFactory.DataInquiry(sensorID, sequenceNumber));

	}
	@Override
	public void sendCustomCommand(int sensorID, int commandID, byte[] values){
		// Stop streaming if we are
		if(isStreaming){
			sendMoteData();
			wasStreaming = true;
		}
		queueSendBytes(CommandFactory.CustomCommand(sensorID, commandID, values, sequenceNumber));
	
		
		
	}

	@Override
	public void disconnect() {
		try {
			in.close();
		} catch (IOException e2) {
			// TODO Auto-generated catch block
			e2.printStackTrace();
		}
		try {
			out.close();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		try {
			sock.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}



	@SuppressLint("UseSparseArrays")
	private void processMoteData(Handler h, byte[] packetData) {
		BANLog.d("PROCESS_PACKET", "Got Mote Data");
		ByteBuffer buff = ByteBuffer.allocate(packetData.length);
		buff.order(ByteOrder.BIG_ENDIAN);
		buff.put(packetData);
		buff.position(0);
		int sensorID;
		HashMap<Integer, ArrayList<Long>> sensorToValues = new HashMap<Integer, ArrayList<Long>>();
		while(buff.hasRemaining()){
			// Figure out the sensorID for this value
			sensorID = buff.get();
			BANLog.d("PROCESS_MOTE_DATA", "Processing data value for sensor " + sensorID);
			ISensor sensor = getSensorById(sensorID);
			if(sensor == null){
				BANLog.e("PROCESS_PACKET", "Got mote data, but not enough information gathered to do anything with it");
				return;

			}
			// Get the valuemappings
			ArrayList<ValueMapping> vm = sensor.getValueMappings();
			// Get the list of values for this sensorID. create a new list if it doesn't yet exist
			ArrayList<Long> rawValues = sensorToValues.get(sensorID);
			if(rawValues == null){
				rawValues = new ArrayList<Long>();
			}
			// Find the right valuemapping for this data value
			for(ValueMapping m : vm){
				if (m.getTypeID() != TYPE_ID.UNSIGNED_INTEGER && m.getTypeID() != TYPE_ID.TWOS_COMPLEMENT) {
					BANLog.e("PROCESS_MOTE_DATA", "Only support UNSIGNED_INTEGERS right now");
					break;
				}
				// Read the raw value
				byte[] rawValue = new byte[m.getSizeInBytes()];
				buff.get(rawValue);
				ByteBuffer valueBuff = ByteBuffer.allocate(8);
				// Pad as needed
				for (int i = 0; i < 8 - m.getSizeInBytes(); i++) {
					valueBuff.put((byte) 0);
				}
				// Extract the raw value as a long (8-byte int)
				valueBuff.put(rawValue);
				valueBuff.position(0);
				long value = valueBuff.getLong();
				// Add this value to the list of values, and save it back in the map
				rawValues.add(value);
			}
			sensorToValues.put(sensorID, rawValues);
		}
		// For each sensor ID...
		for(Integer i : sensorToValues.keySet()){
			Message m = h.obtainMessage();
			Bundle b = new Bundle();
			b.putInt("MESSAGE_TYPE", MESSAGE_TYPE.DATA);
			// Create a simple long array to pass
			ArrayList<Long> results = sensorToValues.get(i);
			long[] values = new long[results.size()];
			for(int j = 0; j < results.size(); j++){
				values[j] = results.get(j);
			}
			b.putLongArray("values", values);
			b.putInt("SensorID", i);
			b.putString("Address", getBluetoothAddr());
			m.setData(b);
			m.sendToTarget();
		}

	}

	private void processDataInquiry(Handler h, byte[] packetData) {
		BANLog.d("PROCESS_PACKET", "Got Data Inquiry Response");
		ByteBuffer buff = ByteBuffer.allocate(packetData.length);
		buff.order(ByteOrder.BIG_ENDIAN);
		buff.put(packetData);
		buff.position(0);
		int sensorID = buff.get();
		BANLog.d("PROCESS_DATA_INQUIRY", "Data inquiry response for sensorID "
				+ sensorID);
		int numValues = buff.get();

		ArrayList<ValueMapping> valueMapping = new ArrayList<ValueMapping>();

		for (int i = 0; i < numValues; i++) {
			int valueSize = buff.get();
			int typeID = buff.get();
			int nameLength = buff.getShort();
			String valueName = readStringFromBytes(buff, nameLength);
			int exprLength = buff.getShort();
			String expression = readStringFromBytes(buff, exprLength);
			int unitsLength = buff.getShort();
			String units = readStringFromBytes(buff, unitsLength);
			valueMapping.add(new ValueMapping(valueSize, typeID, valueName,
					expression, units));
			BANLog.d("PROCESS_DATA_INQUIRY",
					"Added value mapping for sensorID " + sensorID + " name: " + valueName + " eq " + expression);

		}
		ISensor sensor = getSensorById(sensorID);
		sensor.setValueMappings(valueMapping);
		if(numValues > 0){
			Message m = h.obtainMessage();
			Bundle b = new Bundle();
			b.putInt("MESSAGE_TYPE", MESSAGE_TYPE.SENSOR_ADD);
			b.putInt("SensorID", sensorID);
			b.putString("Address", getBluetoothAddr());
			m.setData(b);
			m.sendToTarget();
		}

		if(cascadeLearn){
			if(sensor.getLearnedCommands().size() == 0){
				cascadeLearn = false;
				// Resume streaming if we were
				if(wasStreaming){
					sendMoteData();
					wasStreaming = false;
				}	
			}
			for(ILearnedCommand c : sensor.getLearnedCommands()){
				sendCommandParamsInquiry(sensorID, c.getID());
			}
		}

	}

	private void processCommandInquiry(Handler h, byte[] packetData) {
		BANLog.d("PROCESS_PACKET", "Got Cmd Inquiry Response");
		ByteBuffer buff = ByteBuffer.allocate(packetData.length);
		buff.order(ByteOrder.BIG_ENDIAN);
		buff.put(packetData);
		buff.position(0);
		int sensorID = buff.get();
		BANLog.d("PROCESS_COMMAND_INQUIRY",
				"Command inquiry response for sensorID " + sensorID);
		int numCommands = buff.get();
		ArrayList<ILearnedCommand> commands = new ArrayList<ILearnedCommand>();
		for (int i = 0; i < numCommands; i++) {
			int commandID = buff.get();
			BANLog.d("PROCESS_COMMAND_INQUIRY", "commandID " + commandID);
			short commandNameLength = buff.getShort();
			BANLog.d("PROCESS_COMMAND_INQUIRY", "commandNameLength " + commandNameLength);
			String commandName = readStringFromBytes(buff, commandNameLength);
			BANLog.d("PROCESS_COMMAND_INQUIRY", "Added command " + commandName + " for sensorID "
					+ sensorID);
			commands.add(new LearnedCommand(getSensorById(sensorID), commandID, commandName));	
		}
		if(cascadeLearn){
			sendDataInquiry(sensorID);
			BANLog.d("CASCADE_LEARN", "Continuing to learn about sensorID " + sensorID);

		}
		ISensor sensor = getSensorById(sensorID);
		sensor.setLearnedCommands(commands);

	}

	private void processCommandReturnsInquiry(Handler h, byte[] packetData) {
		BANLog.d("PROCESS_PACKET", "Got Cmd Returns Inquiry Response");
		ByteBuffer buff = ByteBuffer.allocate(packetData.length);
		buff.order(ByteOrder.BIG_ENDIAN);
		buff.put(packetData);
		buff.position(0);
		int sensorID = buff.get();
		int commandID = buff.get();
		int numReturns = buff.get();
		ArrayList<ReturnMapping> returnMapping = new ArrayList<ReturnMapping>();
		for (int i = 0; i < numReturns; i++) {
			int returnSize = buff.get();
			int typeID = buff.get();
			int nameLength = buff.getShort();
			String valueName = readStringFromBytes(buff, nameLength);
			int exprLength = buff.getShort();
			String expression = readStringFromBytes(buff, exprLength);
			int unitsLength = buff.getShort();
			String units = readStringFromBytes(buff, unitsLength);
			returnMapping.add(new ReturnMapping(returnSize, typeID, valueName,
					expression, units));
			BANLog.d("PROCESS_COMMAND_RETURNS",
					"Added return mapping for sensorID " + sensorID
					+ " command " + commandID);

		}
		ISensor sensor = getSensorById(sensorID);
		ILearnedCommand command = sensor.getCommandById(commandID);
		command.setReturnMapping(returnMapping);
		if(cascadeLearn){
			//sendMoteData();
			cascadeLearn = false;
			// Resume streaming if we were
			if(wasStreaming){
				sendMoteData();
				wasStreaming = false;
			}
		}
	}

	private void processCommandParamsInquiry(Handler h, byte[] packetData) {
		BANLog.d("PROCESS_PACKET", "Got Cmd Params Inquiry Response");
		ByteBuffer buff = ByteBuffer.allocate(packetData.length);
		buff.order(ByteOrder.BIG_ENDIAN);
		buff.put(packetData);
		buff.position(0);
		int sensorID = buff.get();
		int commandID = buff.get();
		int numParams = buff.get();
		ArrayList<ParamMapping> paramMapping = new ArrayList<ParamMapping>();
		for (int i = 0; i < numParams; i++) {
			int paramSize = buff.get();
			int typeID = buff.get();
			int nameLength = buff.getShort();
			String paramName = readStringFromBytes(buff, nameLength);
			int restrictionLength = buff.getShort();
			String restriction = readStringFromBytes(buff, restrictionLength);
			int unitsLength = buff.getShort();
			String units = readStringFromBytes(buff, unitsLength);
			paramMapping.add(new ParamMapping(paramSize, typeID, paramName,
					restriction, units));
			BANLog.d("PROCESS_COMMAND_PARAMS", "Added param for sensorID "
					+ sensorID + " command " + commandID);
		}
		ISensor sensor = getSensorById(sensorID);
		ILearnedCommand command = sensor.getCommandById(commandID);
		command.setParams(paramMapping);
		if(cascadeLearn){
			sendCommandReturnsInquiry(sensorID, commandID);
		}
	}


	private void processSensorInquiry(Handler h, byte[] packetData) {
		BANLog.d("PROCESS_PACKET", "Got Sensor Inquiry Response");
		sensorMap = SensorFactory.generateSensorMap(packetData, this);
		if(cascadeLearn){
			for(Integer i : sensorMap.keySet()){
				BANLog.d("CASCADE_LEARN", "Continuing to learn about sensorID " + i);
				sendCommandInquiry(i);

			}
		}

	}

	private void processLearnedCommand(Handler h, byte[] packetData) {
		BANLog.d("PROCESS_LEARNED_COMMAND", "Got ack for sensor " + packetData[0] + " command " + packetData[1]);
		// Redraw GUI, something might have changed
		Message m = h.obtainMessage();
		m.getData().putInt("MESSAGE_TYPE", MESSAGE_TYPE.REDRAW);
		m.sendToTarget();
		
	}



	private String readStringFromBytes(ByteBuffer buff, int numChars) {
		StringBuilder paramName = new StringBuilder();
		for (int j = 0; j < numChars; j++) {
			if(!(buff.position() < (buff.limit()))){
				BANLog.e("READ_STRING_BYTES", "Mismatch between String Length and Availible Bytes in Buffer");
				break;
			}
			paramName.append((char) buff.get());
		}
		return paramName.toString();
	}


	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.wpi.banmqp.shimmer.IShimmer#sendBytes(byte[])
	 */
	private synchronized void queueSendBytes(byte[] bytes){
		messageQueue.add(bytes);
		BANLog.d("SEND_BYTES", "Queued " + Arrays.toString(bytes));
	}

	public synchronized boolean sendNextMsg(){
		byte[] msg = messageQueue.poll();
		if(msg != null && msg.length > 0){
			try {
				out.write(msg);
				BANLog.d("SEND_BYTES", "Sent " + Arrays.toString(msg));
				cts = false;
				return true;
			} catch (IOException e) {
				BANLog.d("SEND_BYTES", "Error Sending " + Arrays.toString(msg));
			}
		}

		return false;
	}

	public boolean clearToSend(){
		return cts;
	}



	@Override
	public boolean waitingToSend() {
		if(messageQueue.peek() != null){
			return true;
		}
		return false;
	}



	@Override
	public boolean isStreaming() {
		return isStreaming;
	}



	@Override
	public void resumeStreaming(boolean b) {
		wasStreaming = b;
		
	}


}
