package edu.wpi.banmqp.shimmer.commands;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import edu.wpi.banmqp.shimmer.MESSAGE_ID;
import edu.wpi.banmqp.shimmer.SequenceNumber;
import edu.wpi.banmqp.util.BANLog;

public abstract class CommandFactory {
	
	private final static int HEADER_SIZE = 5;
	
	/**
	 * Helper method to wrap data with header
	 *
	 * @param messageID the message id
	 * @param packetData the packet data to be wrapped
	 * @param sn the sequencenumber object for this mote
	 * @return the byte[] containing the message to be sent
	 */
	private static byte[] wrapWithHeader(int messageID, byte[] packetData, SequenceNumber sn){
		ByteBuffer buff = ByteBuffer.allocate(HEADER_SIZE + packetData.length);
		buff.order(ByteOrder.BIG_ENDIAN);
		buff.putShort((short) (HEADER_SIZE + packetData.length));
		buff.putShort(sn.next());
		buff.put((byte) messageID);
		buff.put(packetData);
		return buff.array();
	}
	
	/**
	 * Sensor inquiry.
	 *
	 * @param sn the sequencenumber object for this mote
	 * @return the byte[] for a sensor inquiry message
	 */
	public static byte[] SensorInquiry(SequenceNumber sn){
		return wrapWithHeader(MESSAGE_ID.SENSOR_INQUIRY, new byte[0], sn);
	}
	
	/**
	 * Command inquiry.
	 *
	 * @param sensorID the sensor id
	 * @param sn sequencenumber for this mote
	 * @return the byte[] for a command inquiry message
	 */
	public static byte[] CommandInquiry(int sensorID, SequenceNumber sn){
		ByteBuffer buff = ByteBuffer.allocate(1);
		buff.order(ByteOrder.BIG_ENDIAN);
		buff.put((byte) sensorID);
		return wrapWithHeader(MESSAGE_ID.COMMAND_INQUIRY, buff.array(), sn);
	}
	
	/**
	 * Command params inquiry.
	 *
	 * @param sensorID the sensor id
	 * @param commandID the command id
	 * @param sn the sequencenumber for this mote
	 * @return the byte[] for a command params message
	 */
	public static byte[] CommandParamsInquiry(int sensorID, int commandID, SequenceNumber sn){
		ByteBuffer buff = ByteBuffer.allocate(2);
		buff.order(ByteOrder.BIG_ENDIAN);
		buff.put((byte) sensorID);
		buff.put((byte) commandID);
		return wrapWithHeader(MESSAGE_ID.COMMAND_PARAMS_INQUIRY, buff.array(), sn);
	}
	
	/**
	 * Command returns inquiry.
	 *
	 * @param sensorID the sensor id
	 * @param commandID the command id
	 * @param sn the sequencenumber for this mote
	 * @return the byte[] for a command returns message
	 */
	public static byte[] CommandReturnsInquiry(int sensorID, int commandID, SequenceNumber sn){
		ByteBuffer buff = ByteBuffer.allocate(2);
		buff.order(ByteOrder.BIG_ENDIAN);
		buff.put((byte) sensorID);
		buff.put((byte) commandID);
		return wrapWithHeader(MESSAGE_ID.COMMAND_RETURNS_INQUIRY, buff.array(), sn);
	}
	
	/**
	 * Data inquiry.
	 *
	 * @param sensorID the sensor id
	 * @param sn sequencenumber for this mote
	 * @return the byte[] for a data inquiry message
	 */
	public static byte[] DataInquiry(int sensorID, SequenceNumber sn){
		ByteBuffer buff = ByteBuffer.allocate(1);
		buff.order(ByteOrder.BIG_ENDIAN);
		buff.put((byte) sensorID);
		return wrapWithHeader(MESSAGE_ID.DATA_INQUIRY, buff.array(), sn);
	}

	public static byte[] MoteData(SequenceNumber sn) {
		return wrapWithHeader(MESSAGE_ID.MOTE_DATA, new byte[0], sn);
	}

	public static byte[] CustomCommand(int sensorID, int commandID,
			byte[] values, SequenceNumber sn) {
		ByteBuffer buff = ByteBuffer.allocate(2 + values.length);
		buff.order(ByteOrder.BIG_ENDIAN);
		buff.put((byte) sensorID);
		buff.put((byte) commandID);
		buff.put(values);
		return wrapWithHeader(MESSAGE_ID.LEARNED_COMMAND, buff.array(), sn);
	}
	
	
	

}
